
from hypothesis import given
import hypothesis.strategies as st

import strategies

class TopologicalInvariant(object):
    @given(st.data())
    def test_topological_invariants(self, data):
        strategy = getattr(strategies, self._strategy_name)
        
        item = data.draw(strategy())
        h = data.draw(strategies.encodings(item.triangulation))
        image = h(item)
        for method_name in dir(item):
            method = getattr(item, method_name)
            if hasattr(method, 'topological_invariant'):
                image_method = getattr(image, method_name)
                item_property = method()
                image_property = image_method()
                self.assertEqual(item_property, image_property, msg='In %s: %s != %s' % (method_name, item_property, image_property))





''' A module of data structures. '''

from collections import defaultdict, namedtuple
from itertools import chain, islice

import curver

class UnionFind(object):
    ''' A fast union--find data structure. Given items must be hashable. '''
    def __init__(self, items):
        self.parent = dict((item, item) for item in items)
        self.rank = dict((item, 0) for item in items)
    def __iter__(self):
        ''' Iterate through the groups of self. '''
        groups = defaultdict(list)
        for item in self.parent:
            groups[self(item)].append(item)
        return iter(groups.values())
    def __len__(self):
        return sum(1 if self.parent[item] == item else 0 for item in self.parent)
    def __repr__(self):
        return str(self)
    def __str__(self):
        return ', '.join('{' + ', '.join(str(item) for item in g) + '}' for g in self)
    def __call__(self, x):
        ''' Find the root of x. Two items are in the same group iff they have the same root. '''
        root = x
        while self.parent[root] != root:
            root = self.parent[root]
        while self.parent[x] != root:
            x, self.parent[x] = self.parent[x], root
        return root
    def union2(self, x, y):
        ''' Combine the class containing x and the class containing y. '''
        rx, ry = self(x), self(y)
        if self.rank[x] > self.rank[y]:
            self.parent[ry] = rx
        elif self.rank[x] < self.rank[y]:
            self.parent[rx] = ry
        elif rx != ry:
            self.parent[ry] = rx
            self.rank[rx] += 1
    def union(self, *args):
        ''' Combine all of the classes containing the given items. '''
        if len(args) == 1: args = args[0]
        for item in args:
            self.union2(args[0], item)


Terminal = namedtuple('Terminal', ['value'])

class StraightLineProgram(object):
    ''' This represents a straight line program. '''
    def __init__(self, data):
        if isinstance(data, StraightLineProgram):  # Copy.
            data = data.graph
        elif isinstance(data, (list, tuple)):  # Wrap.
            if not data or any(not isinstance(children, (list, tuple)) for children in data) or any(not isinstance(child, (Terminal, curver.IntegerType)) for children in data for child in children):
                data = [[Terminal(child) for child in data]]
        else:
            raise ValueError('Unknown data.')
        
        self.graph = [tuple(children) for children in data]
        self.sinks = [item.value for lst in self.graph for item in lst if isinstance(item, Terminal)]  # !?!
        
        # If w is a descendant of v then w appears before v in self.indices.
        self.indices = []
        used = set()
        
        def dfs(v):
            ''' Perform a depth first traversal starting at the given index. '''
            
            for child in self(v):
                if not isinstance(child, Terminal) and child not in used:
                    dfs(child)
            used.add(v)
            self.indices.append(v)
        dfs(0)
        
        self.num_children = [None] * self.size()
        for index in self.indices:
            self.num_children[index] = sum(1 if isinstance(item, Terminal) else self.num_children[item] for item in self(index))
    
    def __str__(self):
        if len(self) <= 7:
            return str(list(self))
        else:
            return '[%s, %s, %s, ..., %s, %s, %s]' % tuple(chain(islice(self, 3), reversed(list(islice(reversed(self), 3)))))
        
    def __repr__(self):
        strn = []
        for index, item in enumerate(self.graph):
            strn.append('%d --> %s' % (index, item))
        
        return '\n'.join(strn)
    
    def __call__(self, item):
        return self.graph[item]
    def __len__(self):
        return self.num_children[0]
    def size(self):
        ''' Return the number of nodes in this graph. '''
        
        return len(self.graph)
    
    def __getitem__(self, value):
        if isinstance(value, slice):
            # TODO: 2) Implement this.
            return NotImplemented
        else:  # We are returning a single item.
            if value >= len(self) or value < -len(self):
                raise IndexError('index out of range')
            if value < 0: value = len(self) + value
            
            index = 0
            while True:
                for image in self(index):
                    if isinstance(image, Terminal):
                        if value == 0:
                            return image.value
                        else:
                            value -= 1
                    else:
                        if self.num_children[image] > value:
                            index = image
                            break
                        else:
                            value -= self.num_children[image]
        
        raise RuntimeError('Should not be able to reach here.')
    
    def __iter__(self):
        todo = [0]
        while todo:
            v = todo.pop()
            if isinstance(v, Terminal):
                yield v.value
            else:  # isinstance(v, curver.IntegerType):
                todo.extend(reversed(self(v)))
        return
    
    def __lshift__(self, index):
        return [[item if isinstance(item, Terminal) else item + index for item in lst] for lst in self.graph]
    def __rshift__(self, index):
        return [[item if isinstance(item, Terminal) else item - index for item in lst] for lst in self.graph]
    
    def __add__(self, other):
        return StraightLineProgram([[1, self.size()+1]] + (self << 1) + (other << self.size()+1))
    def __radd__(self, other):
        return StraightLineProgram([[1, other.size()+1]] + (other << 1) + (self << other.size()+1))
    
    def __mul__(self, other):
        if other == 0: return StraightLineProgram([])
        
        binary = [bool(int(x)) for x in bin(other)[2:]]
        binary_graph = [[i+2, i+2] for i in range(len(binary)-1)]
        binary_expansion = [i+1 for i, b in enumerate(binary) if b]
        
        return StraightLineProgram([binary_expansion] + binary_graph + (self << len(binary_graph)+1))
    def __rmul__(self, other):
        return self * other
    
    def __contains__(self, value):
        return Terminal(value) in self.sinks
    
    def reverse(self):
        ''' Return the StraightLineProgram that returns self[::-1]. '''
        
        return StraightLineProgram([lst[::-1] for lst in self.graph])
    def __reversed__(self):
        return iter(self.reverse())
    
    def map(self, function=lambda x: x):
        ''' Return the StraightLineProgram obtained by mapping the values of this one under the given function. '''
        
        return StraightLineProgram([[Terminal(function(child.value)) if isinstance(child, Terminal) else child for child in children] for children in self.graph])





''' A module for representing laminations on Triangulations. '''

from itertools import permutations

import curver
from curver.kernel.decorators import memoize, topological_invariant  # Special import needed for decorating.

class Lamination(object):
    ''' This represents an (integral) lamination on a triangulation.
    
    Users should create these via Triangulation(...) or Triangulation.lamination(...). '''
    def __init__(self, triangulation, geometric):
        assert isinstance(triangulation, curver.kernel.Triangulation)
        
        self.triangulation = triangulation
        self.zeta = self.triangulation.zeta
        self.geometric = geometric
        
        # Store some additional weights that are often used.
        self._dual = dict()
        self._side = dict()
        for triangle in self.triangulation:
            i, j, k = triangle  # Edges.
            a, b, c = self.geometric[i.index], self.geometric[j.index], self.geometric[k.index]
            a, b, c = max(a, 0), max(b, 0), max(c, 0)  # Correct for negatives.
            correction = min(a + b - c, b + c - a, c + a - b, 0)
            assert (a + b + c + correction) % 2 == 0
            self._dual[i] = self._side[k] = (b + c - a + correction) // 2
            self._dual[j] = self._side[i] = (c + a - b + correction) // 2
            self._dual[k] = self._side[j] = (a + b - c + correction) // 2
    
    def __repr__(self):
        return '%s(%r, %r)' % (self.__class__.__name__, self.triangulation, self.geometric)
    def __str__(self):
        return '%s %s on %s' % (self.__class__.__name__, self.geometric, self.triangulation)
    def __iter__(self):
        return iter(self.geometric)
    def __call__(self, edge):
        ''' Return the geometric measure assigned to item. '''
        if isinstance(edge, curver.IntegerType): edge = curver.kernel.Edge(edge)  # If given an integer instead.
        
        return self.geometric[edge.index]
    def __eq__(self, other):
        return self.triangulation == other.triangulation and self.geometric == other.geometric
    def __ne__(self, other):
        return not self == other
    def __hash__(self):
        return hash(tuple(self.geometric))
    def __add__(self, other):
        # Haken sum.
        if isinstance(other, Lamination):
            if other.triangulation != self.triangulation:
                raise ValueError('Laminations must be on the same triangulation to add them.')
            
            geometric = [x + y for x, y in zip(self.geometric, other.geometric)]
            return self.triangulation(geometric)  # Have to promote.
        else:
            return NotImplemented
    def __radd__(self, other):
        return self + other  # Commutative.
    def __mul__(self, other):
        assert isinstance(other, curver.IntegerType)
        assert other >= 0
        
        if other == 0:
            new_class = Lamination
        elif other == 1:
            new_class = self.__class__
        elif isinstance(other, curver.kernel.MultiArc):  # or Arc.
            new_class = curver.kernel.MultiArc
        elif isinstance(other, curver.kernel.MultiCurve):  # or Curve.
            new_class = curver.kernel.MultiCurve
        else:
            new_class = Lamination
        
        geometric = [other * x for x in self]
        # TODO: 3) Could save components if they have already been computed.
        return new_class(self.triangulation, geometric)  # Preserve promotion.
    def __rmul__(self, other):
        return self * other  # Commutative.
    
    def weight(self):
        ''' Return the geometric intersection of this lamination with its underlying triangulation. '''
        
        return sum(max(weight, 0) for weight in self)
    
    def dual_weight(self, edge):
        ''' Return the number of component of this lamination dual to the given edge.
        
        Note that when there is a terminal normal arc then we record this weight with a negative sign. '''
        
        if isinstance(edge, curver.IntegerType): edge = curver.kernel.Edge(edge)  # If given an integer instead.
        
        return self._dual[edge]
    
    def side_weight(self, edge):
        ''' Return the number of component of this lamination dual to the given edge.
        
        Note that when there is a terminal normal arc then we record this weight with a negative sign. '''
        
        if isinstance(edge, curver.IntegerType): edge = curver.kernel.Edge(edge)  # If given an integer instead.
        
        return self._side[edge]
    
    @topological_invariant
    def is_empty(self):
        ''' Return if this lamination has no components. '''
        
        return not any(self)  # self.num_components() == 0
    
    @topological_invariant
    def is_multicurve(self):
        ''' Return if this lamination is actually a multicurve. '''
        
        return not self.is_empty() and all(isinstance(component, curver.kernel.Curve) for component in self.components())
    
    @topological_invariant
    def is_curve(self):
        ''' Return if this lamination is actually a curve. '''
        
        return self.is_multicurve() and self.num_components() == 1
    
    @topological_invariant
    def is_multiarc(self):
        ''' Return if this lamination is actually a multiarc. '''
        
        return not self.is_empty() and all(isinstance(component, curver.kernel.Arc) for component in self.components())
    
    @topological_invariant
    def is_arc(self):
        ''' Return if this lamination is actually a multiarc. '''
        
        return self.is_multiarc() and self.num_components() == 1
    
    @topological_invariant
    def is_peripheral(self):
        ''' Return whether this lamination consists entirely of parallel components. '''
        
        return self.peripheral() == self
    
    def promote(self):
        ''' Return this lamination in its finest form. '''
        
        if self.is_multicurve():
            if self.is_curve():
                other = curver.kernel.Curve(self.triangulation, self.geometric)
            else:
                other = curver.kernel.MultiCurve(self.triangulation, self.geometric)
        elif self.is_multiarc():
            if self.is_arc():
                other = curver.kernel.Arc(self.triangulation, self.geometric)
            else:
                other = curver.kernel.MultiArc(self.triangulation, self.geometric)
        else:
            other = self
        
        # Move cache across.
        try:
            other._cache = self._cache  # pylint: disable=attribute-defined-outside-init
        except AttributeError:
            pass  # No cache.
        
        return other
    
    def skeleton(self):
        ''' Return the lamination obtained by collapsing parallel components. '''
        
        return self.triangulation.sum(self.components())
    
    def peek_component(self):
        ''' Return one component of this Lamination. '''
        
        return next(iter(self.components()))
    
    def intersection(self, lamination):
        ''' Return the geometric intersection number between this lamination and the given one. '''
        
        assert isinstance(lamination, Lamination)
        assert lamination.triangulation == self.triangulation
        
        conjugator = self.shorten()
        short = conjugator(self)
        short_lamination = conjugator(lamination)
        
        intersection = 0
        # Parallel components.
        for component, (multiplicity, p) in short.parallel_components().items():
            if isinstance(component, curver.kernel.Arc):
                intersection += multiplicity * max(short_lamination(p), 0)
            else:  # isinstance(component, curver.kernel.Curve):
                v = short.triangulation.vertex_lookup[p]  # = self.triangulation.vertex_lookup[~p].
                v_edges = curver.kernel.utilities.cyclic_slice(v, p, ~p)  # The set of edges that come out of v from p round to ~p.
                
                around_v = min(max(short_lamination.side_weight(edge), 0) for edge in v_edges)
                out_v = sum(max(-short_lamination.side_weight(edge), 0) for edge in v_edges) + sum(max(-short_lamination(edge), 0) for edge in v_edges[1:])
                # around_v > 0 ==> out_v == 0; out_v > 0 ==> around_v == 0.
                intersection += multiplicity * (max(short_lamination(p), 0) - 2 * around_v + out_v)
        
        # Peripheral components.
        for _, (multiplicity, vertex) in short.peripheral_components().items():
            intersection += multiplicity * sum(max(-short_lamination(edge), 0) + max(-short_lamination.side_weight(edge), 0) for edge in vertex)
        
        return intersection
    
    def no_common_component(self, lamination):
        ''' Return that self does not share any components with the given Lamination. '''
        
        assert isinstance(lamination, Lamination)
        
        self_components = self.components()
        return not any(component in self_components for component in lamination.components())
    
    @topological_invariant
    def num_components(self):
        ''' Return the total number of components. '''
        
        return sum(self.components().values())
    
    def sublaminations(self):
        ''' Return all sublaminations that appear within self. '''
        
        components = self.components()
        return [self.triangulation.sum(sub) for i in range(len(components)) for sub in permutations(components, i+1)]  # Powerset.
    
    def peripheral(self, promote=True):
        ''' Return the lamination consisting of the peripheral components of this Lamination. '''
        
        geometric = [0] * self.zeta
        for component, (multiplicity, _) in self.peripheral_components().items():
            geometric = [x + multiplicity * y for x, y in zip(geometric, component)]
        
        return self.triangulation(geometric, promote)  # Have to promote.
    
    def non_peripheral(self, promote=True):
        ''' Return the lamination consisting of the non-peripheral components of this Lamination. '''
        
        geometric = [x - y for x, y in zip(self, self.peripheral(promote=False))]
        
        return self.triangulation(geometric, promote)  # Have to promote.
    
    def multiarc(self):
        ''' Return the maximal MultiArc contained within this lamination. '''
        
        return self.triangulation.sum([multiplicity * component for component, multiplicity in self.components().items() if isinstance(component, curver.kernel.Arc)])
    
    def multicurve(self):
        ''' Return the maximal MultiCurve contained within this lamination. '''
        
        return self.triangulation.sum([multiplicity * component for component, multiplicity in self.components().items() if isinstance(component, curver.kernel.Curve)])
    
    def boundary(self):
        ''' Return the boundary of a regular neighbourhood of this lamination. '''
        
        if self.is_empty():
            return self
        
        return self.multiarc().boundary() + self.multicurve().boundary()
    
    @topological_invariant
    def is_filling(self):
        ''' Return if this Lamination fills the surface, that is, if it intersects all curves on the surface.
        
        Note that this is equivalent to:
            - it meets every non S_{0,3} component of the surface, and
            - its boundary is peripheral.
        
        Furthermore, if any component of this lamination is a non-peripheral curve then it cannot fill. '''
        
        if any(isinstance(component, curver.kernel.Curve) and not component.is_peripheral() for component in self.components()):
            return False
        
        for component in self.triangulation.components():
            V, E = len([vertex for vertex in self.triangulation.vertices if vertex[0] in component]), len(component) // 2
            if (V, E) != (3, 3):  # component != S_{0, 3}:
                if all(self(edge) == 0 for edge in component):
                    return False
        
        return self.boundary().is_peripheral()
    
    def fills_with(self, other):  # pylint: disable=no-self-use
        ''' Return whether self \\cup other fills. '''
        assert isinstance(other, Lamination)
        
        return NotImplemented  # TODO: 2) Implement! (And remove the PyLint disable when done.)
    
    def trace(self, edge, intersection_point, length):
        ''' Return the sequence of edges encountered by following along this lamination.
        
        We start at the given edge and intersection point for the specified number of steps.
        However we terminate early if the lamination closes up before this number of steps is completed. '''
        
        if isinstance(edge, curver.IntegerType): edge = curver.kernel.Edge(edge)  # If given an integer instead.
        
        start = (edge, intersection_point)
        
        assert 0 <= intersection_point < self(edge)  # Sanity.
        dual_weights = dict((edge, self.dual_weight(edge)) for edge in self.triangulation.edges)
        edges = []
        for _ in range(length):
            x, y, z = self.triangulation.corner_lookup[~edge]
            if intersection_point < dual_weights[z]:  # Turn right.
                edge, intersection_point = y, intersection_point
            elif dual_weights[x] < 0 and dual_weights[z] <= intersection_point < dual_weights[z] - dual_weights[x]:  # Terminate.
                break
            else:  # Turn left.
                edge, intersection_point = z, self(z) - self(x) + intersection_point
            if (edge, intersection_point) == start:  # Closes up.
                break
            edges.append(edge)
            assert 0 <= intersection_point < self(edge)  # Sanity.
        
        return edges
    
    # @topological_invariant
    def topological_type(self):  # pylint: disable=no-self-use
        ''' Return the topological type of this lamination..
        
        Two laminations are in the same mapping class group orbit if and only their topological types are equal.
        These are labelled graphs and so equal means 'label isomorphic', so we return a custom class that uses networkx.is_isomorphic to determine equality. '''
        
        # This should generalise MultiCurve.topological_type().
        # The final code should be very similar but with some additional steps:
        #  6) Let a := crush(self.multiarc()).
        #  7) Label each vertex with the topological type of a restricted to the corresponding component.
        # Note that in order for this to work the topological type MUST be compatible with the permutation of the punctures.
        
        return NotImplemented  # TODO: 2) Implement. (And remove the PyLint disable when done.)
    
    def peripheral_components(self):
        ''' Return a dictionary mapping component |--> (multiplicity, vertex) for each component of self that is peripheral around a vertex. '''
        
        components = dict()
        for vertex in self.triangulation.vertices:
            multiplicity = min(max(self.side_weight(edge), 0) for edge in vertex)
            if multiplicity > 0:
                component = self.triangulation.curve_from_cut_sequence(vertex)
                components[component] = (multiplicity, vertex)
        
        return components
    
    @memoize
    def parallel_components(self):
        ''' Return a dictionary mapping component |--> (multiplicity, edge) for each component of self that is parallel to an edge. '''
        
        components = dict()
        for edge in self.triangulation.edges:
            if edge.sign() == +1:  # Don't double count.
                multiplicity = -self(edge)
                if multiplicity > 0:
                    component = self.triangulation.edge_arc(edge)
                    components[component] = (multiplicity, edge)
            
            if self.triangulation.vertex_lookup[edge] == self.triangulation.vertex_lookup[~edge]:
                v = self.triangulation.vertex_lookup[edge]  # = self.triangulation.vertex_lookup[~edge].
                v_edges = curver.kernel.utilities.cyclic_slice(v, edge, ~edge)  # The set of edges that come out of v from edge round to ~edge.
                if len(v_edges) > 2:
                    around_v = min(max(self.side_weight(edge), 0) for edge in v_edges)
                    twisting = min(max(self.side_weight(edge) - around_v, 0) for edge in v_edges[1:-1])
                    
                    if self.side_weight(v_edges[0]) == around_v and self.side_weight(v_edges[-1]) == around_v:
                        multiplicity = twisting
                        
                        if multiplicity > 0:
                            component = self.triangulation.curve_from_cut_sequence(v_edges[1:])
                            components[component] = (multiplicity, edge)
        
        return components
    
    @memoize
    def components(self):
        ''' Return a dictionary mapping components to their multiplicities. '''
        
        components = dict()
        
        conjugator = self.shorten()
        short = conjugator(self)
        conjugator_inv = conjugator.inverse()
        
        for component, (multiplicity, _) in short.peripheral_components().items():
            components[conjugator_inv(component)] = multiplicity
        for component, (multiplicity, _) in short.parallel_components().items():
            components[conjugator_inv(component)] = multiplicity
        
        return components
    
    def is_short(self):
        ''' Return whether this lamination is short.
        
        A lamination is short if all of its non-peripheral components are parallel to edges of
        the triangulation that it is defined on. This makes computing its components very easy. '''
        
        lamination = self.non_peripheral(promote=False)
        
        geometric = list(lamination)
        for component, (multiplicity, _) in lamination.parallel_components().items():
            geometric = [x - y * multiplicity for x, y in zip(geometric, component)]
        lamination = Lamination(lamination.triangulation, geometric)
        
        return lamination.is_empty()
    
    @memoize
    def shorten(self):
        ''' Return an encoding which maps this lamination to a short one. '''
        
        lamination = self.non_peripheral(promote=False)
        conjugator = self.triangulation.id_encoding()
        
        if self.is_short():  # If this lamination is already short:
            return conjugator
        
        def shorten_strategy(self, edge):
            ''' Return a float in [0, 1] describing how good flipping this edge is for making this lamination short. '''
            
            if isinstance(edge, curver.IntegerType): edge = curver.kernel.Edge(edge)  # If given an integer instead.
            
            if not self.triangulation.is_flippable(edge): return 0
            
            ad, bd, cd, dd, ed = [self.dual_weight(edgy) for edgy in self.triangulation.square(edge)]
            
            if ed < 0 or (ed == 0 and ad > 0 and bd > 0):  # Non-parallel arc or bipod.
                return 1
            
            return 0
        
        def curve_shorten_strategy(self, edge):
            ''' Return a float in [0, 1] describing how good flipping this edge is for making this curve short. '''
            
            if isinstance(edge, curver.IntegerType): edge = curver.kernel.Edge(edge)  # If given an integer instead.
            
            if not self.triangulation.is_flippable(edge): return 0
            
            ai, bi, ci, di, ei = [self(edgy) for edgy in self.triangulation.square(edge)]
            ad, bd, cd, dd, ed = [self.dual_weight(edgy) for edgy in self.triangulation.square(edge)]
            
            if ei == 0:
                return 0
            if max(ai + ci, bi + di) - ei < ei:
                return 1  # Hmmm. We do need this but can we avoid not having 1 as the generic case?
            if bd > 0 and ad == 0:
                return 0.75
            
            return 0.5
        
        active_edges = set(lamination.triangulation.edges)  # Edges that are not currently parallel to a component and so can be flipped.
        while not lamination.is_empty():
            
            extra = []  # High priority edges to check.
            while active_edges:
                # Could be another try / except block.
                edge = curver.kernel.utilities.maximum(extra + list(active_edges), key=lambda edge: shorten_strategy(lamination, edge), upper_bound=1)
                if shorten_strategy(lamination, edge) == 0: break  # No non-parallel arcs or bipods.
                
                a, b, c, d, e = lamination.triangulation.square(edge)
                flip = lamination.triangulation.encode_flip(edge)  # edge is always flippable.
                try:  # Accelerate!
                    if (1 - 0.1 / lamination.zeta) * lamination.weight() > flip(lamination).weight():  # Drop is at least 10% of average edge weight.
                        raise curver.AssumptionError('Flip made definite progress.')
                    
                    intersection_point = lamination.side_weight(e) if lamination.side_weight(e) > 0 else -lamination.dual_weight(a)
                    trace = lamination.trace(edge, intersection_point, 2*self.zeta)
                    trace = trace[:trace.index(edge)+1]  # Will raise a ValueError if edge is not in trace.
                    
                    curve = lamination.triangulation.lamination_from_cut_sequence(trace)
                    if not isinstance(curve, curver.kernel.Curve):
                        raise ValueError
                    
                    slope = curve.slope(lamination)  # Will raise a curver.AssumptionError if these are disjoint.
                    if -1 <= slope <= 1:  # Can't accelerate. We should probably also skip cases where slope is too close to small to be efficient.
                        raise ValueError
                    else:  # slope < -1 or 1 < slope:
                        move = curve.encode_twist(power=-int(slope))  # Round towards zero.
                except (ValueError, curver.AssumptionError):
                    move = flip
                    extra = [x for x in [c, d] if x in active_edges]
                
                conjugator = move * conjugator
                lamination = move(lamination)
                if lamination(edge) <= 0:
                    active_edges.discard(edge)
                    active_edges.discard(~edge)
            
            # There should now be no bigons and all arcs should be parallel to edges.
            assert all(lamination.side_weight(edge) >= 0 for edge in lamination.triangulation.edges)
            assert all([lamination.side_weight(edge) > 0 for edge in triangle].count(True) != 2 for triangle in lamination.triangulation)
            
            # This is pretty inefficient.
            for edge in active_edges:
                if lamination(edge) > 0 and lamination.side_weight(edge) == 0:
                    trace = [edge] + lamination.trace(edge, 0, 2*self.zeta)
                    curve = lamination.triangulation.curve_from_cut_sequence(trace)
                    extra = []
                    while not curve.is_short():
                        edgey = curver.kernel.utilities.maximum(extra + list(active_edges), key=lambda edge: curve_shorten_strategy(curve, edge), upper_bound=1)
                        # This edge is always flippable.
                        a, b, c, d, e = lamination.triangulation.square(edgey)
                        
                        move = lamination.triangulation.encode_flip(edgey)
                        extra = [x for x in [c, d] if x in active_edges]
                        conjugator = move * conjugator
                        curve = move(curve)
                        lamination = move(lamination)
            
            # Subtract.
            geometric = list(lamination)
            for component, (multiplicity, edge) in lamination.parallel_components().items():
                if lamination(edge) <= 0:
                    geometric = [x - y * multiplicity for x, y in zip(geometric, component)]
                    active_edges.discard(edge)
                    active_edges.discard(~edge)
            
            lamination = Lamination(lamination.triangulation, geometric)
        
        assert conjugator(self).is_short()  # Sanity.
        
        return conjugator





import hypothesis.strategies as st
from hypothesis.stateful import Bundle, RuleBasedStateMachine, rule

import curver

class UnionFindRules(RuleBasedStateMachine):
    def __init__(self):
        super(UnionFindRules, self).__init__()
        self.initialize([])
    
    @rule(items=st.sets(elements=st.integers()))
    def initialize(self, items):
        self.__items = list(items)
        self.__union_find = curver.kernel.UnionFind(items)
    
    @rule(data=st.data())
    def find(self, data):
        a = data.draw(st.sampled_from(self.__items))
        self.__union_find(a)
    
    @rule(data=st.data())
    def iterate(self, data):
        assert set(self.__items) == set(sum(self.__union_find, []))
    
    @rule(data=st.data())
    def union2(self, data):
        a = data.draw(st.sampled_from(self.__items))
        b = data.draw(st.sampled_from(self.__items))
        self.__union_find.union2(a, b)
        assert self.__union_find(a) == self.__union_find(b)
    
    @rule(data=st.data())
    def union(self, data):
        items = data.draw(st.lists(elements=st.sampled_from(self.__items)))
        self.__union_find.union(items)
        for item in items:
            assert self.__union_find(item) == self.__union_find(items[0])

TestUnionFind = UnionFindRules.TestCase

class SLPRules(RuleBasedStateMachine):
    SLPs = Bundle('slps')
    
    @rule(target=SLPs, items=st.lists(elements=st.integers()))
    def newslp(self, items):
        return curver.kernel.StraightLineProgram(items)
    
    @rule(target=SLPs, slp=SLPs)
    def copy(self, slp):
        copy = curver.kernel.StraightLineProgram(slp)
        assert list(copy) == list(slp)
        return copy
    
    @rule(target=SLPs, slp1=SLPs, slp2=SLPs)
    def add(self, slp1, slp2):
        added = slp1 + slp2
        assert list(added) == list(slp1) + list(slp2)
        return added
    
    @rule(slp=SLPs, factor=st.integers(min_value=0, max_value=1000))
    def multiply(self, slp, factor):
        assert list(slp * factor) == list(slp) * factor
        assert list(factor * slp) == list(slp * factor)
    
    @rule(data=st.data())
    def getitem(self, data):
        slp = data.draw(self.SLPs.filter(lambda s: len(s) > 0))  # Non-empty.
        index = data.draw(st.integers(min_value=0, max_value=len(slp)-1))
        assert list(slp)[index] == slp[index]
    
    @rule(target=SLPs, slp=SLPs)
    def reverse(self, slp):
        rev = slp.reverse()
        assert list(rev) == list(slp)[::-1]
        assert list(reversed(slp)) == list(slp)[::-1]
        return rev

TestSLP = SLPRules.TestCase





from hypothesis import given
import hypothesis.strategies as st
import pytest
import unittest

import curver

# TRIANGULATIONS = {(g, p): curver.load(g, p).triangulation.sig() for g in range(10) for p in range(1, 10) if 6*g + 3*p - 6 >= 3}
TRIANGULATIONS = {
    (0, 3): '3_p',
    (0, 4): '6_JDky1',
    (0, 5): '9_ZFO4OuIV',
    (0, 6): 'c_ZlgeM906o-354',
    (0, 7): 'f_ZBhisze7ZhPen+Kdl1',
    (0, 8): 'i_ZBCol3oodJti4fabgfrLdu1',
    (0, 9): 'l_ZBmFdZ5TmXJTHbErcDHjdC12vbv4',
    (1, 1): '3_o',
    (1, 2): '6_WKSv',
    (1, 3): '9_q3AtQIBp',
    (2, 1): '9_qBFcQEN4',
    (2, 2): 'c_q5YHALgCWTyw',
    (2, 3): 'f_q5sG87f308pRczx4k',
    (3, 1): 'f_0mGaGVParMxLYnCS3',
    (3, 2): 'i_0mGG3b5sZ48hMGl+XMxQf7',
    }

SIGNATURES = [TRIANGULATIONS[key] for key in sorted(TRIANGULATIONS)]

@st.composite
def triangulations(draw):
    sig = draw(st.sampled_from(SIGNATURES))
    return curver.triangulation_from_sig(sig)

MCGS = [curver.load(g, p) for g, p in [(0, 3), (0, 4), (1, 1), (1, 2), (1, 3), (2, 1), (2, 2), (2, 3), (3, 1), (3, 2)]]

@st.composite
def mcgs(draw):
    return draw(st.sampled_from(MCGS))

@st.composite
def mapping_classes(draw, mcg=None):
    if mcg is None: mcg = draw(mcgs())
    return mcg(draw(st.integers(min_value=0, max_value=10)))

@st.composite
def encodings(draw, triangulation=None):
    if triangulation is None: triangulation = draw(triangulations())
    rev_sequence = [triangulation.id_isometry()]
    num_flips = draw(st.integers(min_value=0, max_value=20))
    for _ in range(num_flips):
        T = rev_sequence[-1].target_triangulation
        if draw(st.sampled_from([0, 0, 0, 0, 0, 1])) == 0:
            edge = draw(st.sampled_from([edge for edge in T.edges if T.is_flippable(edge)]))
            move = T.encode_flip(edge)[0]
        else:
            move = T.encode_relabel_edges([i if draw(st.booleans()) else ~i for i in draw(st.permutations(range(T.zeta)))])[0]
        rev_sequence.append(move)
    
    return curver.kernel.Encoding(rev_sequence[::-1])

@st.composite
def homology_classes(draw, triangulation=None):
    if triangulation is None: triangulation = draw(triangulations())
    algebraic = [draw(st.integers()) for _ in range(triangulation.zeta)]
    return curver.kernel.HomologyClass(triangulation, algebraic)

@st.composite
def multiarcs(draw, triangulation=None, require_non_empty_boundary=False):
    if triangulation is None: triangulation = draw(triangulations())
    
    geometric = [0] * triangulation.zeta
    
    available_indices = set(triangulation.indices)
    num_arcs = draw(st.integers(min_value=1, max_value=len(available_indices)))
    for _ in range(num_arcs):
        index = draw(st.sampled_from(sorted(available_indices)))
        available_indices.remove(index)
        geometric[index] = draw(st.integers(max_value=-1))
    
    return triangulation.lamination(geometric)

@st.composite
def arcs(draw, triangulation=None):
    return draw(multiarcs(triangulation)).peek_component()

@st.composite
def multicurves(draw, triangulation=None):
    if triangulation is None: triangulation = draw(triangulations())
    
    # Build a special type of multiarc with non-empty boundary.
    geometric = [0] * triangulation.zeta
    
    indices = set()
    available_indices = set(triangulation.indices)
    
    tree = triangulation.dual_tree()
    tree = set([index for index in triangulation.indices if tree[index]])
    available_indices = available_indices - tree
    
    if draw(st.booleans()) and not any(g == 0 for (g, v) in triangulation.surface()):  # merge vertices:
        classes = curver.kernel.UnionFind(triangulation.vertices)
        for index in sorted(available_indices):
            arc = triangulation.edge_arc(index)
            if arc.connects_distinct_vertices():
                a, b = arc.vertices()
                if classes(a) != classes(b):
                    classes.union(a, b)
                    indices.add(index)
                    available_indices.remove(index)
    
    num_arcs = draw(st.integers(min_value=0 if indices else 1, max_value=len(available_indices)-1))
    for _ in range(num_arcs):
        index = draw(st.sampled_from(sorted(available_indices)))
        available_indices.remove(index)
        indices.add(index)
    
    # Set weights on these edges.
    for index in indices:
        geometric[index] = draw(st.integers(max_value=-1))
    
    multiarc = triangulation.lamination(geometric)
    boundary = multiarc.boundary()
    assert(not boundary.is_empty())
    return boundary

@st.composite
def curves(draw, triangulation=None):
    multicurve = draw(multicurves(triangulation))
    curve = multicurve.peek_component()
    return curve

@st.composite
def laminations(draw, triangulation=None):
    return draw(st.one_of(multicurves(triangulation), multiarcs(triangulation)))

@st.composite
def permutations(draw, N=None):
    if N is None: N = draw(st.integers(min_value=1, max_value=10))
    return curver.kernel.Permutation(draw(st.permutations(range(N))))


class TestStrategiesHealth(unittest.TestCase):
    @given(triangulations())
    def test_triangulations(self, triangulation):
        self.assertIsInstance(triangulation, curver.kernel.Triangulation)
    
    @given(mcgs())
    def test_mcgs(self, mcg):
        self.assertIsInstance(mcg, curver.kernel.MappingClassGroup)
    
    @given(mapping_classes())
    def test_mapping_classes(self, h):
        self.assertIsInstance(h, curver.kernel.MappingClass)
    
    @given(encodings())
    def test_encodings(self, encoding):
        self.assertIsInstance(encoding, curver.kernel.Encoding)
    
    @given(homology_classes())
    def test_homology_classes(self, hc):
        self.assertIsInstance(hc, curver.kernel.HomologyClass)
    
    @given(arcs())
    def test_arcs(self, arc):
        self.assertIsInstance(arc, curver.kernel.Arc)
    
    @given(multiarcs())
    def test_multiarcs(self, multiarc):
        self.assertIsInstance(multiarc, curver.kernel.MultiArc)
    
    @given(curves())
    def test_curves(self, curve):
        self.assertIsInstance(curve, curver.kernel.Curve)
    
    @given(multicurves())
    def test_multicurves(self, multicurve):
        self.assertIsInstance(multicurve, curver.kernel.MultiCurve)
    
    @pytest.mark.skip('Incomplete')
    @given(laminations())
    def test_laminations(self, lamination):
        self.assertIsInstance(lamination, curver.kernel.Lamination)
    
    @given(permutations())
    def test_permutations(self, perm):
        self.assertIsInstance(perm, curver.kernel.Permutation)


if __name__ == '__main__':
    unittest.main()





''' A module for representing more advanced ways of changing triangulations. '''

import curver
from curver.kernel.moves import Move  # Special import needed for subclassing.

class Crush(Move):
    ''' This represents the effect of crushing along a curve. '''
    def __init__(self, source_triangulation, target_triangulation, curve, matrix):
        super(Crush, self).__init__(source_triangulation, target_triangulation)
        
        assert isinstance(curve, curver.kernel.Curve)
        assert not curve.is_peripheral() and curve.is_short()
        assert curve.triangulation == self.source_triangulation
        
        self.curve = curve
        self.matrix = matrix
    
    def __str__(self):
        return 'Crush ' + str(self.curve)
    def __reduce__(self):
        return (self.__class__, (self.source_triangulation, self.target_triangulation, self.curve, self.matrix))
    def __eq__(self, other):
        if isinstance(other, Crush):
            return self.source_triangulation == other.source_triangulation and self.target_triangulation == other.target_triangulation and self.curve == other.curve and self.matrix == other.matrix
        else:
            return NotImplemented
    def __ne__(self, other):
        return not self == other
    
    def apply_lamination(self, lamination):
        geometric = list(lamination)
        
        # Get some edges.
        a = self.curve.parallel()
        _, b, e = self.source_triangulation.corner_lookup[a]
        
        v = self.curve.triangulation.vertex_lookup[a]  # = self.triangulation.vertex_lookup[~a].
        v_edges = curver.kernel.utilities.cyclic_slice(v, a, ~a)  # The set of edges that come out of v from a round to ~a.
        around_v = min(max(lamination.side_weight(edge), 0) for edge in v_edges)
        out_v = sum(max(-lamination.side_weight(edge), 0) for edge in v_edges) + sum(max(-lamination(edge), 0) for edge in v_edges[1:])
        # around_v > 0 ==> out_v == 0; out_v > 0 ==> around_v == 0.
        twisting = min(max(lamination.side_weight(edge) - around_v, 0) for edge in v_edges[1:-1])
        # Computing around_v and twisting can be done more efficiently.
        
        # We work by manipulating the side weights around v.
        sides = dict((edge, lamination.side_weight(edge) - (self.curve.side_weight(edge)*twisting + around_v if edge in v_edges and lamination.side_weight(edge) >= 0 else 0)) for edge in self.source_triangulation.edges)
        parallels = dict((edge.index, max(-lamination(edge), 0)) for edge in v_edges)
        
        # TODO: 4) Add comments explaining what is going on in the next two blocks and how the different tightening cases work.
        
        # Tighten to the left.
        drop = max(sides[a], 0) + max(-sides[b], 0)
        for edge in v_edges[1:-1]:
            x, y, z = lamination.triangulation.corner_lookup[edge]
            if sides[x] >= 0 and sides[y] >= 0 and sides[z] >= 0:
                if drop <= sides[x]:
                    sides[x] = sides[x] - drop
                else:  # sides[x] < drop.
                    sides[x], sides[y], drop = sides[x] - drop, sides[y] + sides[x] - drop, sides[x]
            elif sides[x] < 0:
                sides[x], sides[y], drop = sides[x] - drop, sides[y] - drop, 0
            elif sides[y] < 0:
                sides[x] = sides[x] - drop
            else:  # sides[z] < 0.
                if drop <= sides[x]:
                    sides[x] = sides[x] - drop
                elif sides[x] < drop <= sides[x] - sides[z]:
                    parallels[z.index] = parallels[z.index] + (drop - sides[x])
                    sides[x], sides[z], drop = 0, sides[z] + (drop - sides[x]), sides[x]
                else:  # sides[x] - sides[z] < drop:
                    parallels[z.index] = parallels[z.index] - sides[z]
                    sides[x], sides[y], sides[z], drop = sides[x] - sides[z] - drop, sides[y] - (drop - sides[x] + sides[z]), 0, sides[x]
            
            if drop == 0: break  # Stop early.
        
        # Tighten to the right.
        drop = max(-sides[a], 0) + max(sides[b], 0)
        for edge in reversed(v_edges[1:-1]):
            x, y, z = lamination.triangulation.corner_lookup[edge]
            if sides[x] >= 0 and sides[y] >= 0 and sides[z] >= 0:
                if drop <= sides[x]:
                    sides[x] = sides[x] - drop
                else:  # sides[x] < drop.
                    sides[x], sides[z], drop = sides[x] - drop, sides[z] + sides[x] - drop, sides[x]
            elif sides[x] < 0:
                sides[x], sides[z], drop = sides[x] - drop, sides[z] - drop, 0
            elif sides[y] < 0:
                if drop <= sides[x]:
                    sides[x] = sides[x] - drop
                elif sides[x] < drop <= sides[x] - sides[y]:
                    parallels[x.index] = parallels[x.index] + (drop - sides[x])
                    sides[x], sides[y], drop = 0, sides[y] + (drop - sides[x]), sides[x]
                else:  # sides[x] - sides[y] < drop:
                    parallels[x.index] = parallels[x.index] - sides[y]
                    sides[x], sides[y], sides[z], drop = sides[x] - sides[y] - drop, 0, sides[z] - (drop - sides[x] + sides[y]), sides[x]
            else:  # sides[z] < 0.
                sides[x] = sides[x] - drop
            
            if drop == 0: break  # Stop early.
        
        # Now rebuild the intersection.
        for edge in v_edges:
            if edge not in (a, b, e, ~b, ~e):
                x, y, z = lamination.triangulation.corner_lookup[edge]
                if parallels[edge.index] > 0:
                    geometric[edge.index] = -parallels[x.index]
                else:
                    geometric[edge.index] = max(sides[x], 0) + max(sides[y], 0) + max(-sides[z], 0)
                    
                    # Sanity check:
                    x2, y2, z2 = lamination.triangulation.corner_lookup[~edge]
                    assert geometric[edge.index] == max(sides[x2], 0) + max(sides[y2], 0) + max(-sides[z2], 0)
        
        # We have to rebuild the ~e edge separately since it now pairs with ~b.
        x, y, z = lamination.triangulation.corner_lookup[~e]
        if parallels[e.index] + parallels[b.index] + max(-sides[e], 0) > 0:
            geometric[e.index] = -(parallels[e.index] + parallels[b.index] + max(-sides[e], 0))
        else:
            geometric[e.index] = max(sides[x], 0) + max(sides[y], 0) + max(-sides[z], 0)
            
            # Sanity check:
            x2, y2, z2 = lamination.triangulation.corner_lookup[~b]
            assert geometric[e.index] == max(sides[x2], 0) + max(sides[y2], 0) + max(-sides[z2], 0)
        
        # And finally the b edge, which is now paired with e.
        # Since around_v > 0 ==> out_v == 0 & out_v > 0 ==> around_v == 0, this is equivalent to: around_v if around_v > 0 else -out_v
        geometric[b.index] = around_v - out_v
        
        return self.target_triangulation(geometric)  # Have to promote.
    
    def apply_homology(self, homology_class):
        return NotImplemented  # I don't think we ever need this.
    
    def inverse(self):
        return Lift(self.target_triangulation, self.source_triangulation, self.curve, self.matrix)

class Lift(Move):
    ''' This represents the inverse of crushing along a curve. '''
    def __init__(self, source_triangulation, target_triangulation, curve, matrix):
        super(Lift, self).__init__(source_triangulation, target_triangulation)
        
        assert isinstance(curve, curver.kernel.Curve)
        assert curve.triangulation == self.target_triangulation
        
        self.curve = curve
        self.matrix = matrix
    
    def __str__(self):
        return 'Lift ' + str(self.curve)
    def __reduce__(self):
        return (self.__class__, (self.source_triangulation, self.target_triangulation, self.curve, self.matrix))
    def __eq__(self, other):
        if isinstance(other, Lift):
            return self.source_triangulation == other.source_triangulation and self.target_triangulation == other.target_triangulation and self.curve == other.curve and self.matrix == other.matrix
        else:
            return NotImplemented
    def __ne__(self, other):
        return not self == other
    
    def apply_lamination(self, lamination):
        # Really should check that the dual weights around a vertex are all non-negative.
        geometric = curver.kernel.matrix_vector(self.matrix, lamination.geometric)
        return lamination.__class__(self.target_triangulation, geometric)  # Avoid promote since the lift has to be the same type as the given lamination.
    
    def apply_homology(self, homology_class):
        return NotImplemented  # I don't think we ever need this.
    
    def inverse(self):
        return Crush(self.target_triangulation, self.source_triangulation, self.curve, self.matrix)





''' A module of useful, generic functions; including input and output formatting. '''

from itertools import product
from string import ascii_lowercase, ascii_uppercase, digits
import re

import curver

ALPHABET = digits + ascii_lowercase + ascii_uppercase + '+-'

def b64encode(n):
    ''' Return n in base 64. '''
    
    out = []
    while n:
        n, r = divmod(n, 64)
        out.append(ALPHABET[r])
    return ''.join(out)

def b64decode(strn):
    ''' Return the integer with base 64 encoding strn. '''
    
    return sum(ALPHABET.index(c) * 64**i for i, c in enumerate(strn))

def string_generator(n, skip=None):
    ''' Return a list of n usable names, none of which are in skip. '''
    
    assert isinstance(n, curver.IntegerType)
    assert skip is None or isinstance(skip, (list, tuple, dict, set))
    
    skip = set() if skip is None else set(skip)
    
    alphabet = ascii_lowercase
    results = []
    i = 0
    while len(results) < n:
        i += 1
        for letters in product(alphabet, repeat=i):
            word = ''.join(letters)
            if word not in skip:
                results.append(word)
            if len(results) >= n: break
    
    return results

def name_objects(objects, skip=None):
    ''' Return a list of pairs (name, object). '''
    
    assert isinstance(objects, (list, tuple))
    
    return zip(string_generator(len(objects), skip), objects)

def cyclic_slice(L, x, y):
    ''' Return the sublist of L from x (inclusive) to y (exclusive).
    
    L may be cyclically permuted if needed. '''
    i = L.index(x)
    L = L[i:] + L[:i]  # x is now L[0].
    j = L.index(y)
    return L[:j]

def maximum(iterable, key=lambda x: x, upper_bound=None):
    ''' Return the maximum of iterable but terminate early when given an upper_bound. '''

    def helper():
        ''' A generator that yeilds elements from iterable up to and including one such that key(item) >= upper_bound. '''
        for item in iterable:
            yield item
            if upper_bound is not None and key(item) >= upper_bound: return

    return max(helper(), key=key)

def alphanum_key(strn):
    ''' Return a list of string and number chunks from a string. '''
    
    blocks = []
    for chunk in re.split('([0-9]+)', strn):
        try:
            blocks.append(int(chunk))
        except ValueError:
            blocks.append(chunk)
    
    return blocks

def matrix_vector(M, v):
    ''' Return M*v where v is interpreted as a column vector. '''
    
    return (v * M.T).tolist()[0]





''' A module for decorators. '''

from functools import wraps

def memoize(function):
    ''' A decorator that memoizes a method of a class. '''
    @wraps(function)
    def caching(self):
        ''' The cached version of function.
        
        Note that this docstring will be overwritten with functions docstring by the wraps decorator. '''
        if not hasattr(self, '_cache'):
            self._cache = dict()
        key = function.__name__
        if key not in self._cache:
            self._cache[key] = function(self)
        return self._cache[key]
    
    return caching

def topological_invariant(function):
    ''' Mark this function as a topological invariant.
    
    This is allows it to be picked out by the TopologicalInvariant unittests. '''
    
    function.topological_invariant = True
    return function




from hypothesis import given
import hypothesis.strategies as st
from string import ascii_lowercase
import unittest

import curver

class TestMisc(unittest.TestCase):
    @given(st.integers(min_value=0))
    def test_b64(self, integer):
        strn = curver.kernel.utilities.b64encode(integer)
        self.assertEqual(curver.kernel.utilities.b64decode(strn), integer)

    @given(st.integers(min_value=0, max_value=1000), st.sets(elements=st.text(ascii_lowercase)))
    def test_string_generator(self, n, skip):
        strns = curver.kernel.utilities.string_generator(n, skip)
        self.assertEqual(len(strns), n)
        self.assertFalse(skip.intersection(strns))
    
    @given(st.data())
    def test_maximum(self, data):
        bound = data.draw(st.integers())
        integers = data.draw(st.lists(elements=st.integers(max_value=bound), min_size=1))
        value = curver.kernel.utilities.maximum(integers, upper_bound=bound)
        self.assertEqual(value, min(max(integers), bound))

